from __future__ import annotations

import unittest

import numpy as np
import pandas as pd

import prince


class TestGPA(unittest.TestCase):
    def setUp(self):
        # Create a list of 2-D circles with different locations and rotations
        n_shapes = 4
        n_points = 12
        n_dims = 2

        shape_sizes = np.arange(1, n_shapes + 1)
        shape_angle_offsets = 10 * np.arange(n_shapes)
        shape_center_offsets = np.tile(np.arange(n_shapes), (n_dims, 1))

        base_angles = np.linspace(0, 2 * np.pi, num=n_points, endpoint=False)
        # Size (n_shapes, n_points)
        angles = base_angles[np.newaxis, :] + shape_angle_offsets[:, np.newaxis]

        # Calculate along dimensions
        x = np.cos(angles) * shape_sizes[:, np.newaxis] + shape_center_offsets[0][:, np.newaxis]
        y = np.sin(angles) * shape_sizes[:, np.newaxis] + shape_center_offsets[1][:, np.newaxis]

        self.shapes = np.stack([x, y], axis=-1)

    def test_fit(self):
        gpa = prince.GPA()
        self.assertIsInstance(gpa.fit(self.shapes), prince.GPA)

    def test_fit_random(self):
        gpa = prince.GPA(init="random")
        self.assertIsInstance(gpa.fit(self.shapes), prince.GPA)

    def test_fit_mean(self):
        gpa = prince.GPA(init="mean")
        self.assertIsInstance(gpa.fit(self.shapes), prince.GPA)

    def test_fit_bad_init(self):
        gpa = prince.GPA(init="bad init type")

        with self.assertRaises(ValueError):
            gpa.fit(self.shapes)

    def test_fit_bad_input_size(self):
        gpa = prince.GPA()

        with self.assertRaises(ValueError):
            gpa.fit(self.shapes[0])

    def test_transform(self):
        gpa = prince.GPA(copy=True)
        aligned_shapes = gpa.fit(self.shapes).transform(self.shapes)
        self.assertIsInstance(aligned_shapes, np.ndarray)
        self.assertEqual(self.shapes.shape, aligned_shapes.shape)

    def test_fit_transform_equal(self):
        """In our specific case of all-same-shape circles, the shapes should
        align perfectly."""
        gpa = prince.GPA()
        aligned_shapes = gpa.fit_transform(self.shapes)
        self.assertIsInstance(aligned_shapes, np.ndarray)
        np.testing.assert_array_almost_equal(aligned_shapes[:-1], aligned_shapes[1:])

    def test_fit_transform_single(self):
        """Aligning a single shape should return the same shape, just normalized."""
        gpa = prince.GPA()
        shapes = self.shapes[0:1]
        aligned_shapes = gpa.fit_transform(shapes)
        np.testing.assert_array_almost_equal(shapes / np.linalg.norm(shapes), aligned_shapes)

    def test_copy(self):
        shapes_copy = np.copy(self.shapes)

        gpa = prince.GPA(copy=True)
        gpa.fit(shapes_copy)
        np.testing.assert_array_equal(self.shapes, shapes_copy)

        gpa = prince.GPA(copy=False)
        gpa.fit(shapes_copy)
        self.assertRaises(AssertionError, np.testing.assert_array_equal, self.shapes, shapes_copy)

    def test_xarray(self):
        points = pd.DataFrame(
            data=[
                [0, 0, 0, 0],
                [0, 2, 0, 1],
                [1, 0, 0, 2],
                [3, 2, 1, 0],
                [1, 2, 1, 1],
                [3, 3, 1, 2],
                [0, 0, 2, 0],
                [0, 4, 2, 1],
                [2, 0, 2, 2],
            ],
            columns=["x", "y", "shape", "point"],
        ).astype({"x": float, "y": float})

        ds = points.set_index(["shape", "point"]).to_xarray()
        da = ds.to_stacked_array("xy", ["shape", "point"])
        shapes = da.values

        gpa = prince.GPA()
        aligned_shapes = gpa.fit_transform(shapes)
        da.values = aligned_shapes
        da.to_unstacked_dataset("xy").to_dataframe().reset_index()



from __future__ import annotations

import math
import tempfile

import numpy as np
import pandas as pd
import pytest
import rpy2.robjects as robjects
import sklearn.utils.estimator_checks
import sklearn.utils.validation
from rpy2.robjects import r as R

import prince
from tests import load_df_from_R


@pytest.mark.parametrize(
    "sup_rows, sup_groups",
    [
        pytest.param(sup_rows, sup_groups, id=f"{sup_rows=}:{sup_groups=}")
        for sup_rows in [False, True]
        for sup_groups in [False, True]
    ],
)
class TestMFA:
    _row_name = "row"
    _col_name = "col"

    @pytest.fixture(autouse=True)
    def _prepare(self, sup_rows, sup_groups):
        self.sup_rows = sup_rows
        self.sup_groups = sup_groups

        n_components = 3

        # Fit Prince
        self.dataset = prince.datasets.load_premier_league()
        active = self.dataset.copy()
        if self.sup_rows:
            active = active.drop(index=["Manchester City", "Manchester United"])
        supplementary_groups = ["2023-24"] if self.sup_groups else []
        self.groups = self.dataset.columns.levels[0].tolist()
        self.mfa = prince.MFA(n_components=n_components)
        self.mfa.fit(active, groups=self.groups, supplementary_groups=supplementary_groups)

        # Fit FactoMineR
        R("library('FactoMineR')")
        with tempfile.NamedTemporaryFile() as fp:
            dataset = self.dataset.copy()
            dataset.columns = [" ".join(parts) for parts in dataset.columns]
            dataset.to_csv(fp, index=False)
            R(f"dataset <- read.csv('{fp.name}')")

        args = "dataset, group=c(6, 6, 6), graph=F"
        if self.sup_rows:
            args += ", ind.sup=c(9:10)"
        if self.sup_groups:
            args += ", num.group.sup=c(3)"

        R(f"mfa <- MFA({args})")

    def test_check_is_fitted(self):
        assert isinstance(self.mfa, prince.MFA)
        sklearn.utils.validation.check_is_fitted(self.mfa)

    def test_total_inertia(self):
        F = robjects.r("sum(mfa$eig[,1])")[0]
        P = self.mfa.total_inertia_
        assert math.isclose(F, P)

    def test_eigenvalues(self):
        F = load_df_from_R("mfa$eig")[: self.mfa.n_components]
        P = self.mfa._eigenvalues_summary
        np.testing.assert_allclose(F["eigenvalue"], P["eigenvalue"])
        np.testing.assert_allclose(F["percentage of variance"], P["% of variance"])
        np.testing.assert_allclose(
            F["cumulative percentage of variance"], P["% of variance (cumulative)"]
        )

    def test_group_eigenvalues(self):
        for i, group in enumerate(self.groups, start=1):
            F = load_df_from_R(f"mfa$separate.analyses$Gr{i}$eig")[: self.mfa.n_components]
            P = self.mfa[group]._eigenvalues_summary
            np.testing.assert_allclose(F["eigenvalue"], P["eigenvalue"])
            np.testing.assert_allclose(F["percentage of variance"], P["% of variance"])
            np.testing.assert_allclose(
                F["cumulative percentage of variance"], P["% of variance (cumulative)"]
            )

    @pytest.mark.parametrize("method_name", ("row_coordinates", "transform"))
    def test_row_coords(self, method_name):
        method = getattr(self.mfa, method_name)
        F = load_df_from_R("mfa$ind$coord")
        P = method(self.dataset)
        if self.sup_rows:
            F = pd.concat((F, load_df_from_R("mfa$ind.sup$coord")))
            # Move supplementary rows to the end
            P = pd.concat(
                [
                    P.loc[P.index.difference(["Manchester City", "Manchester United"])],
                    P.loc[["Manchester City", "Manchester United"]],
                ]
            )
        F = F.iloc[:, : self.mfa.n_components]
        np.testing.assert_allclose(F.abs(), P.abs())

    def test_row_contrib(self):
        F = load_df_from_R("mfa$ind$contrib").iloc[:, : self.mfa.n_components]
        P = self.mfa.row_contributions_
        np.testing.assert_allclose(F, P * 100)



from __future__ import annotations

import math

import numpy as np
import pandas as pd
import pytest
import rpy2.robjects as robjects
import sklearn.utils.estimator_checks
import sklearn.utils.validation
from rpy2.robjects import numpy2ri
from sklearn import decomposition, pipeline, preprocessing

import prince
from tests import load_df_from_R


@pytest.mark.parametrize(
    "sup_rows, sup_cols, scale, sample_weights, column_weights",
    [
        pytest.param(
            sup_rows,
            sup_cols,
            scale,
            sample_weights,
            column_weights,
            id=f"{sup_rows=}:{sup_cols=}:{scale=}:{sample_weights=}:{column_weights=}",
        )
        for sup_rows in [False, True]
        for sup_cols in [False, True]
        for scale in [False, True]
        for sample_weights in [False, True]
        for column_weights in [False, True]
    ],
)
class TestPCA:
    @pytest.fixture(autouse=True)
    def _prepare(self, sup_rows, sup_cols, scale, sample_weights, column_weights):
        self.sup_rows = sup_rows
        self.sup_cols = sup_cols
        self.scale = scale

        n_components = 5

        # Fit Prince
        self.dataset = prince.datasets.load_decathlon()
        self.active = self.dataset.copy()
        if self.sup_rows:
            self.active = self.active.query('competition == "Decastar"')
        self.sample_weights = (
            np.random.default_rng().dirichlet([1] * len(self.active)) if sample_weights else None
        )
        supplementary_columns = ["rank", "points"] if self.sup_cols else []
        self.column_weights = (
            np.random.default_rng().random(
                len(self.active.columns.difference(supplementary_columns))
            )
            if column_weights
            else None
        )
        self.pca = prince.PCA(n_components=n_components, rescale_with_std=self.scale)
        self.pca.fit(
            self.active,
            sample_weight=self.sample_weights,
            column_weight=self.column_weights,
            supplementary_columns=supplementary_columns,
        )

        # scikit-learn
        if self.scale:
            self.sk_pca = pipeline.make_pipeline(
                preprocessing.StandardScaler(),
                decomposition.PCA(n_components=n_components),
            )
        else:
            self.sk_pca = pipeline.make_pipeline(
                decomposition.PCA(n_components=n_components),
            )
        # sklearn's PCA doesn't support sample weights
        self.sk_pca.fit(self.active[self.pca.feature_names_in_])

        # Fit FactoMineR
        robjects.r(
            """
        library('FactoMineR')

        data(decathlon)
        decathlon <- subset(decathlon, select = -c(Competition))
        """
        )

        args = f"decathlon, ncp={n_components}, graph=F"
        if sample_weights:
            robjects.r.assign("row.w", numpy2ri.py2rpy(self.sample_weights))
            robjects.r("row.w <- as.vector(row.w)")
            args += ", row.w=row.w"
        if column_weights:
            robjects.r.assign("col.w", numpy2ri.py2rpy(self.column_weights))
            robjects.r("col.w <- as.vector(col.w)")
            args += ", col.w=col.w"
        if not self.scale:
            args += ", scale.unit=F"
        if self.sup_cols:
            if self.sup_rows:
                robjects.r(f"pca = PCA({args}, quanti.sup=c(11, 12), ind.sup=c(14:41))")
            else:
                robjects.r(f"pca = PCA({args}, quanti.sup=c(11, 12))")
        else:
            if self.sup_rows:
                robjects.r(f"pca = PCA({args}, ind.sup=c(14:41))")
            else:
                robjects.r(f"pca = PCA({args})")

    def test_check_is_fitted(self):
        assert isinstance(self.pca, prince.PCA)
        sklearn.utils.validation.check_is_fitted(self.pca)

    def test_total_inertia(self):
        F = robjects.r("sum(pca$eig[,1])")[0]
        P = self.pca.total_inertia_
        assert math.isclose(F, P)

    def test_eigenvalues(self):
        P = self.pca._eigenvalues_summary
        # Test against FactoMineR
        F = load_df_from_R("pca$eig")[: self.pca.n_components]
        np.testing.assert_allclose(F["eigenvalue"], P["eigenvalue"])
        np.testing.assert_allclose(F["percentage of variance"], P["% of variance"])
        np.testing.assert_allclose(
            F["cumulative percentage of variance"], P["% of variance (cumulative)"]
        )
        # Test against scikit-learn
        if self.sample_weights is None and self.column_weights is None:
            n = len(self.active)
            S = self.sk_pca[-1].explained_variance_ * (n - 1) / n
            np.testing.assert_allclose(P["eigenvalue"], S)
            np.testing.assert_allclose(
                P["% of variance"], self.sk_pca[-1].explained_variance_ratio_ * 100
            )

    @pytest.mark.parametrize("method_name", ("row_coordinates", "transform"))
    def test_row_coords(self, method_name):
        method = getattr(self.pca, method_name)
        P = method(self.dataset)
        # Test againt FactoMineR
        F = load_df_from_R("pca$ind$coord")
        if self.sup_rows:
            F = pd.concat((F, load_df_from_R("pca$ind.sup$coord")))
        np.testing.assert_allclose(F.abs(), P.abs())
        # Test against scikit-learn
        if self.sample_weights is None and self.column_weights is None:
            S = self.sk_pca.transform(self.dataset[self.pca.feature_names_in_])
            np.testing.assert_allclose(np.abs(S), P.abs())

    def test_row_cosine_similarities(self):
        F = load_df_from_R("pca$ind$cos2")
        if self.sup_rows:
            F = pd.concat((F, load_df_from_R("pca$ind.sup$cos2")))
        P = self.pca.row_cosine_similarities(self.dataset)
        np.testing.assert_allclose(F, P)

    def test_row_contrib(self):
        F = load_df_from_R("pca$ind$contrib")
        P = self.pca.row_contributions_
        np.testing.assert_allclose(F, P * 100)

    def test_col_coords(self):
        F = load_df_from_R("pca$var$coord")
        P = self.pca.column_coordinates_
        if self.sup_cols:
            P = P.drop(["rank", "points"])
        np.testing.assert_allclose(F.abs(), P.abs())

    def test_col_cos2(self):
        F = load_df_from_R("pca$var$cos2")
        P = self.pca.column_cosine_similarities_
        if self.sup_cols:
            P = P.drop(["rank", "points"])
        np.testing.assert_allclose(F, P)

    def test_col_contrib(self):
        F = load_df_from_R("pca$var$contrib")
        P = self.pca.column_contributions_
        np.testing.assert_allclose(F, P * 100)



"""Multiple Correspondence Analysis (MCA)"""

from __future__ import annotations

import numpy as np
import pandas as pd
import sklearn.base
import sklearn.preprocessing
import sklearn.utils

from prince import utils

from . import ca


class MCA(sklearn.base.BaseEstimator, sklearn.base.TransformerMixin, ca.CA):
    def __init__(
        self,
        n_components=2,
        n_iter=10,
        copy=True,
        check_input=True,
        random_state=None,
        engine="sklearn",
        one_hot=True,
        one_hot_prefix_sep="__",
        one_hot_columns_to_drop=None,
        correction=None,
    ):
        if correction is not None:
            if correction not in {"benzecri", "greenacre"}:
                raise ValueError("correction must be either 'benzecri' or 'greenacre' if provided.")
            if not one_hot:
                raise ValueError(
                    "correction can only be applied when one_hot is True. This is because the "
                    "number of original variables is needed to apply the correction."
                )

        super().__init__(
            n_components=n_components,
            n_iter=n_iter,
            copy=copy,
            check_input=check_input,
            random_state=random_state,
            engine=engine,
        )
        self.one_hot = one_hot
        self.one_hot_prefix_sep = one_hot_prefix_sep
        self.one_hot_columns_to_drop = one_hot_columns_to_drop
        self.correction = correction

    def _prepare(self, X):
        if self.one_hot:
            X = pd.get_dummies(X, columns=X.columns, prefix_sep=self.one_hot_prefix_sep)
            if self.one_hot_columns_to_drop is not None:
                X = X.drop(columns=self.one_hot_columns_to_drop, errors="ignore")
            if (one_hot_columns_ := getattr(self, "one_hot_columns_", None)) is not None:
                X = X.reindex(columns=one_hot_columns_.union(X.columns), fill_value=False)
        return X

    def get_feature_names_out(self, input_features=None):
        return np.arange(self.n_components_)

    @property
    def eigenvalues_(self):
        """Returns the eigenvalues associated with each principal component."""
        eigenvalues = super().eigenvalues_
        # Benzécri and Greenacre corrections
        if self.correction in {"benzecri", "greenacre"}:
            K = self.K_
            return np.array(
                [(K / (K - 1) * (eig - 1 / K)) ** 2 if eig > 1 / K else 0 for eig in eigenvalues]
            )
        return eigenvalues

    @property
    @utils.check_is_fitted
    def percentage_of_variance_(self):
        """Returns the percentage of explained inertia per principal component."""
        # Benzécri correction
        if self.correction == "benzecri":
            eigenvalues = self.eigenvalues_
            return 100 * eigenvalues / eigenvalues.sum()
        # Greenacre correction
        if self.correction == "greenacre":
            eigenvalues = super().eigenvalues_
            benzecris = self.eigenvalues_
            K, J = (self.K_, self.J_)
            average_inertia = (K / (K - 1)) * ((eigenvalues**2).sum() - (J - K) / K**2)
            return 100 * benzecris / average_inertia
        # No correction
        return super().percentage_of_variance_

    @utils.check_is_dataframe_input
    def fit(self, X, y=None):
        """Fit the MCA for the dataframe X.

        The MCA is computed on the indicator matrix (i.e. `X.get_dummies()`). If some of the columns are already
        in indicator matrix format, you'll want to pass in `K` as the number of "real" variables that it represents.
        (That's used for correcting the inertia linked to each dimension.)

        """

        if self.check_input:
            sklearn.utils.check_array(X, dtype=[str, "numeric"])

        # K is the number of actual variables, to apply the Benzécri correction
        self.K_ = X.shape[1]

        # One-hot encode the data
        one_hot = self._prepare(X)
        self.one_hot_columns_ = one_hot.columns

        # We need the number of columns to apply the Greenacre correction
        self.J_ = one_hot.shape[1]

        # Apply CA to the indicator matrix
        super().fit(one_hot)

        return self

    @utils.check_is_dataframe_input
    @utils.check_is_fitted
    def row_coordinates(self, X):
        return super().row_coordinates(self._prepare(X))

    @utils.check_is_dataframe_input
    @utils.check_is_fitted
    def row_cosine_similarities(self, X):
        oh = self._prepare(X)
        return super()._row_cosine_similarities(X=oh, F=super().row_coordinates(oh))

    @utils.check_is_dataframe_input
    @utils.check_is_fitted
    def column_coordinates(self, X):
        return super().column_coordinates(self._prepare(X))

    @utils.check_is_dataframe_input
    @utils.check_is_fitted
    def column_cosine_similarities(self, X):
        oh = self._prepare(X)
        return super()._column_cosine_similarities(X=oh, G=super().column_coordinates(oh))

    @utils.check_is_dataframe_input
    @utils.check_is_fitted
    def transform(self, X):
        """Computes the row principal coordinates of a dataset."""
        if self.check_input:
            sklearn.utils.check_array(X, dtype=[str, "numeric"])
        return self.row_coordinates(X)



from __future__ import annotations

import tempfile

import numpy as np
import pandas as pd
import pytest
from rpy2.robjects import r as R

import prince
from tests import load_df_from_R
from tests.test_ca import TestCA as _TestCA


class TestMCA(_TestCA):
    _row_name = "ind"
    _col_name = "var"

    @pytest.fixture(autouse=True)
    def _prepare(self, sup_rows, sup_cols):
        self.sup_rows = sup_rows
        self.sup_cols = sup_cols

        n_components = 5
        n_active_rows = 1_000

        # Fit Prince
        self.dataset = prince.datasets.load_hearthstone_cards()
        active = self.dataset.copy()
        if self.sup_rows:
            active = active[:n_active_rows]
        if self.sup_cols:
            active = active.drop(columns=["type_or_school"])
        self.ca = prince.MCA(n_components=n_components, engine="scipy")
        self.ca.fit(active)

        # Fit FactoMineR
        R("library('FactoMineR')")
        with tempfile.NamedTemporaryFile() as fp:
            self.dataset.to_csv(fp)
            R(f"dataset <- read.csv('{fp.name}')[,-1]")

        args = f"dataset, ncp={n_components}, graph=F"
        if self.sup_cols:
            if self.sup_rows:
                R(
                    f"ca <- MCA({args}, quali.sup=c(4), ind.sup=c({n_active_rows + 1}:nrow(dataset)))"
                )
            else:
                R(f"ca <- MCA({args}, quali.sup=c(4))")
        else:
            if self.sup_rows:
                R(f"ca <- MCA({args}, ind.sup=c({n_active_rows + 1}:nrow(dataset)))")
            else:
                R(f"ca <- MCA({args})")

    @pytest.mark.parametrize("method_name", ("row_coordinates", "transform"))
    def test_row_coords(self, method_name):
        super().test_row_coords(method_name=method_name)

    def test_col_coords(self):
        if self.sup_cols:
            F = load_df_from_R("ca$var$coord")
            if self.sup_cols:
                F = pd.concat((F, load_df_from_R("ca$quali.sup$coord")))
            P = self.ca.column_coordinates(self.dataset)
            # Prince adds a prefix to each column. We need to remove it in order to align the rows
            # of the two dataframes
            P.index = [idx.split("__", 1)[1] for idx in P.index]
            np.testing.assert_allclose(F.abs(), P.abs().loc[F.index])
        else:
            super().test_col_coords()

    def test_col_cos2(self):
        if self.sup_cols:
            F = load_df_from_R("ca$var$cos2")
            if self.sup_cols:
                F = pd.concat((F, load_df_from_R("ca$quali.sup$cos2")))
            P = self.ca.column_cosine_similarities(self.dataset)
            # Prince adds a prefix to each column. We need to remove it in order to align the rows
            # of the two dataframes
            P.index = [idx.split("__", 1)[1] for idx in P.index]
            np.testing.assert_allclose(F, P.loc[F.index])
        else:
            super().test_col_cos2()


def test_with_and_without_one_hot():
    """

    >>> df = pd.DataFrame({
    ...     "foo": [1, 2, 3, 3, 5],
    ...     "bar": ["a", "b", "c", "b", "e"],
    ... })
    >>> mca = prince.MCA(n_components=2, one_hot=True, engine="scipy")
    >>> mca = mca.fit(df)
    >>> coords = mca.transform(df)
    >>> assert coords.shape == (5, 2)
    >>> coords.round(2).abs().sort_index(axis='columns')  # doctest: +SKIP
          0    1
    0  0.00  2.0
    1  0.65  0.5
    2  0.65  0.5
    3  0.65  0.5
    4  1.94  0.5

    >>> mca = prince.MCA(n_components=2, one_hot=False, engine="scipy")
    >>> one_hot = pd.get_dummies(df, columns=['foo', 'bar'])
    >>> mca = mca.fit(one_hot)
    >>> coords = mca.transform(one_hot)
    >>> assert coords.shape == (5, 2)
    >>> coords.round(2).abs().sort_index(axis='columns')  # doctest: +SKIP
          0    1
    0  0.00  1.0
    1  0.65  0.5
    2  0.65  0.5
    3  0.65  0.5
    4  1.94  0.5

    """


def test_issue_131():
    """

    https://github.com/MaxHalford/prince/issues/131#issuecomment-1591426031

    >>> df = pd.DataFrame({
    ...     "foo": [1, 2, 3, 3, 5],
    ...     "bar": ["a", "b", "c", "b", "e"],
    ... })
    >>> mca = prince.MCA(engine="scipy")
    >>> mca = mca.fit(df)
    >>> coords = mca.transform(df)
    >>> assert coords.shape == (5, 2)
    >>> coords.round(2).abs().sort_index(axis='columns')  # doctest: +SKIP
          0    1
    0  0.00  2.0
    1  0.65  0.5
    2  0.65  0.5
    3  0.65  0.5
    4  1.94  0.5

    >>> mca.K_, mca.J_
    (2, 8)

    """


def test_issue_171():
    """

    https://github.com/MaxHalford/prince/issues/171

    >>> from sklearn import impute
    >>> from sklearn import pipeline

    >>> rng = np.random.RandomState(0)
    >>> test_data = pd.DataFrame(data=rng.random((10, 5)))
    >>> test = pipeline.Pipeline(steps=[
    ...     ('impute', impute.SimpleImputer()),  # would break the pipeline since it returns an ndarray
    ...     ('mca', prince.PCA()),
    ... ])
    >>> _ = test[0].set_output(transform='pandas')
    >>> test.fit_transform(test_data)
    component         0         1
    0         -0.392617  0.296831
    1          0.119661 -1.660653
    2         -1.541581 -0.826863
    3          3.105498 -0.538801
    4         -2.439259 -0.343292
    5          1.129341 -0.533576
    6         -1.077436  0.899673
    7          0.020571 -0.941029
    8          1.498005  1.566376
    9         -0.422184  2.081334

    """


def test_type_doesnt_matter():
    """

    Checks that the type of the columns doesn't affect the result.

    """
    outputs = []
    dataset = prince.datasets.load_hearthstone_cards().head(100)
    for col in dataset.columns:
        labels, levels = pd.factorize(dataset[col])
        dataset[col] = labels
    for typ in ("int", "float", "str", "category"):
        dataset = dataset.astype(typ)
        mca = prince.MCA(n_components=2, engine="scipy")
        mca = mca.fit(dataset)
        outputs.append(mca.transform(dataset).abs())

    for i in range(len(outputs) - 1):
        np.testing.assert_allclose(outputs[i], outputs[i + 1])


issue_161_data = """
,category,userid,location,applicationname,browser\n
0,Portal Login,a@b.com,"San Jose, CA, United States",A,Chrome\n
1,Application Access,b@b.com,"San Jose, CA, United States",B,Other\n
2,Application Access,a@b.com,"San Jose, CA, United States",C,Other\n
3,Portal Login,c@b.com,"San Diego, CA, United States",A,Chrome\n
"""


def test_issue_161():
    """

    https://github.com/MaxHalford/prince/issues/161

    >>> import io
    >>> data = pd.read_csv(io.StringIO(issue_161_data), index_col=0)

    >>> mca = prince.MCA(
    ...     n_components=10,
    ...     n_iter=3,
    ...     copy=True,
    ...     check_input=True,
    ...     engine='sklearn',
    ...     random_state=42
    ... )
    >>> mca = mca.fit(data[:3])

    >>> mca.eigenvalues_summary
              eigenvalue % of variance % of variance (cumulative)
    component
    0              0.673        67.32%                     67.32%
    1              0.327        32.68%                    100.00%

    >>> mca.row_coordinates(data[:3])
              0         1
    0  1.120811 -0.209242
    1 -0.820491 -0.571660
    2 -0.300320  0.780902

    >>> mca.transform(data[3:])
              0         1
    3  1.664888 -0.640285

    """


def test_abdi_2007_correction():
    """

    >>> wines = prince.datasets.load_burgundy_wines()
    >>> wines = wines.drop(columns=["Oak type"], level=0)

    >>> mca = prince.MCA(n_components=4, correction=None)
    >>> mca = mca.fit(wines)
    >>> mca.eigenvalues_.round(4).tolist()
    [0.8532, 0.2, 0.1151, 0.0317]
    >>> mca.percentage_of_variance_.round(3).tolist()
    [71.101, 16.667, 9.593, 2.64]

    >>> mca = prince.MCA(n_components=4, correction="benzecri")
    >>> mca = mca.fit(wines)
    >>> mca.eigenvalues_.round(4).tolist()
    [0.7004, 0.0123, 0.0003, 0.0]
    >>> mca.percentage_of_variance_.round(3).tolist()
    [98.229, 1.731, 0.04, 0.0]

    >>> mca = prince.MCA(n_components=4, correction="greenacre")
    >>> mca = mca.fit(wines)
    >>> mca.eigenvalues_.round(4).tolist()
    [0.7004, 0.0123, 0.0003, 0.0]
    >>> mca.percentage_of_variance_.round(3).tolist()
    [95.189, 1.678, 0.038, 0.0]

    """



from __future__ import annotations

from collections import OrderedDict

import numpy as np
from scipy import linalg

GRAY = OrderedDict([("light", "#bababa"), ("dark", "#404040")])


def stylize_axis(ax, grid=True):
    if grid:
        ax.grid()

    ax.xaxis.set_ticks_position("none")
    ax.yaxis.set_ticks_position("none")

    ax.axhline(y=0, linestyle="-", linewidth=1.2, color=GRAY["dark"], alpha=0.6)
    ax.axvline(x=0, linestyle="-", linewidth=1.2, color=GRAY["dark"], alpha=0.6)

    return ax


def build_ellipse(X, Y):
    """Construct ellipse coordinates from two arrays of numbers.

    Args:
        X (1D array_like)
        Y (1D array_like)

    Returns:
        float: The mean of `X`.
        float: The mean of `Y`.
        float: The width of the ellipse.
        float: The height of the ellipse.
        float: The angle of orientation of the ellipse.

    """
    x_mean = np.mean(X)
    y_mean = np.mean(Y)

    cov_matrix = np.cov(np.vstack((X, Y)))
    U, s, V = linalg.svd(cov_matrix, full_matrices=False)

    chi_95 = np.sqrt(4.61)  # 90% quantile of the chi-square distribution
    width = np.sqrt(cov_matrix[0][0]) * chi_95 * 2
    height = np.sqrt(cov_matrix[1][1]) * chi_95 * 2

    eigenvector = V.T[0]
    angle = np.arctan(eigenvector[1] / eigenvector[0])

    return x_mean, y_mean, width, height, angle



from __future__ import annotations

import tempfile

import numpy as np
import pytest
import sklearn.utils.estimator_checks
import sklearn.utils.validation
from rpy2.robjects import r as R

import prince
from tests import load_df_from_R


@pytest.mark.parametrize(
    "sup_rows, sup_cols",
    [
        pytest.param(
            sup_rows,
            sup_cols,
            id=":".join(["sup_rows" if sup_rows else "", "sup_cols" if sup_cols else ""]).strip(
                ":"
            ),
        )
        for sup_rows in [False]
        for sup_cols in [False]
    ],
)
class TestFAMD:
    _row_name = "row"
    _col_name = "col"

    @pytest.fixture(autouse=True)
    def _prepare(self, sup_rows, sup_cols):
        self.sup_rows = sup_rows
        self.sup_cols = sup_cols

        n_components = 5

        # Fit Prince
        self.dataset = prince.datasets.load_beers().head(200)
        active = self.dataset.copy()
        self.famd = prince.FAMD(n_components=n_components, engine="scipy")
        self.famd.fit(active)

        # Fit FactoMineR
        R("library('FactoMineR')")
        with tempfile.NamedTemporaryFile() as fp:
            self.dataset.to_csv(fp)
            R(f"dataset <- read.csv('{fp.name}', row.names=c(1))")
            R("famd <- FAMD(dataset, graph=F)")

    def test_check_is_fitted(self):
        assert isinstance(self.famd, prince.FAMD)
        sklearn.utils.validation.check_is_fitted(self.famd)

    def test_num_cols(self):
        assert sorted(self.famd.num_cols_) == [
            "alcohol_by_volume",
            "final_gravity",
            "international_bitterness_units",
            "standard_reference_method",
        ]

    def test_cat_cols(self):
        assert sorted(self.famd.cat_cols_) == ["is_organic", "style"]

    def test_eigenvalues(self):
        F = load_df_from_R("famd$eig")[: self.famd.n_components]
        P = self.famd._eigenvalues_summary
        np.testing.assert_allclose(F["eigenvalue"], P["eigenvalue"])
        np.testing.assert_allclose(F["percentage of variance"], P["% of variance"])
        np.testing.assert_allclose(
            F["cumulative percentage of variance"], P["% of variance (cumulative)"]
        )

    @pytest.mark.parametrize("method_name", ("row_coordinates", "transform"))
    def test_row_coords(self, method_name):
        method = getattr(self.famd, method_name)
        F = load_df_from_R("famd$ind$coord")
        P = method(self.dataset)
        np.testing.assert_allclose(F.abs(), P.abs())

    def test_row_contrib(self):
        F = load_df_from_R("famd$ind$contrib")
        P = self.famd.row_contributions_
        np.testing.assert_allclose(F, P * 100)

    def test_col_coords(self):
        F = load_df_from_R("famd$var$coord")
        P = self.famd.column_coordinates_
        np.testing.assert_allclose(F.abs(), P.abs())

    def test_col_contrib(self):
        F = load_df_from_R("famd$var$contrib")
        P = self.famd.column_contributions_
        np.testing.assert_allclose(F, P * 100)


def test_issue_169():
    """

    https://github.com/MaxHalford/prince/issues/169

    >>> import pandas as pd
    >>> from prince import FAMD
    >>> df = pd.DataFrame({'var1':['c', 'a', 'b','c'], 'var2':['x','y','y','z'],'var2': [0.,10.,30.4,0.]})

    >>> famd = FAMD(n_components=2, random_state=42)
    >>> famd = famd.fit(df[:3])

    >>> famd.transform(df[0:3])
    component         0         1
    0         -1.303760 -0.658334
    1         -0.335621  0.981047
    2          1.639381 -0.322713

    >>> famd.transform(df[0:2])
    component         0         1
    0         -1.000920 -0.669274
    1         -0.092001  0.669274

    >>> famd.transform(df[3:]).round(6)
    component         0    1
    3         -0.869173 -0.0

    """



from __future__ import annotations

import math
import tempfile

import numpy as np
import pandas as pd
import pytest
import rpy2.robjects as robjects
import sklearn.utils.estimator_checks
import sklearn.utils.validation
from rpy2.robjects import r as R
from scipy import sparse

import prince
from tests import load_df_from_R


@pytest.mark.parametrize(
    "sup_rows, sup_cols",
    [
        pytest.param(
            sup_rows,
            sup_cols,
            id=":".join(["sup_rows" if sup_rows else "", "sup_cols" if sup_cols else ""]).strip(
                ":"
            ),
        )
        for sup_rows in [False, True]
        for sup_cols in [False, True]
    ],
)
class TestCA:
    _row_name = "row"
    _col_name = "col"

    @pytest.fixture(autouse=True)
    def _prepare(self, sup_rows, sup_cols):
        self.sup_rows = sup_rows
        self.sup_cols = sup_cols

        n_components = 5

        # Fit Prince
        self.dataset = prince.datasets.load_french_elections()
        active = self.dataset.copy()
        if sup_rows:
            active = active.drop("Île-de-France")
        if self.sup_cols:
            active = active.drop(columns=["Abstention", "Blank"])
        self.ca = prince.CA(n_components=n_components)
        self.ca.fit(active)

        # Fit FactoMineR
        R("library('FactoMineR')")
        with tempfile.NamedTemporaryFile() as fp:
            self.dataset.to_csv(fp)
            R(f"dataset <- read.csv('{fp.name}', row.names=1)")

        args = f"dataset, ncp={n_components}, graph=F"
        if self.sup_cols:
            if sup_rows:
                R(f"ca <- CA({args}, col.sup=c(13, 14), row.sup=c(18))")
            else:
                R(f"ca <- CA({args}, col.sup=c(13, 14))")
        else:
            if sup_rows:
                R(f"ca <- CA({args}, row.sup=c(18))")
            else:
                R(f"ca <- CA({args})")

    def test_check_is_fitted(self):
        assert isinstance(self.ca, prince.CA)
        sklearn.utils.validation.check_is_fitted(self.ca)

    def test_svd_U(self):
        F = load_df_from_R("ca$svd$U").to_numpy()
        P = sparse.diags(self.ca.row_masses_.to_numpy() ** -0.5) @ self.ca.svd_.U
        np.testing.assert_allclose(np.abs(F), np.abs(P))

    def test_svd_V(self):
        F = load_df_from_R("ca$svd$V").to_numpy()
        P = sparse.diags(self.ca.col_masses_.to_numpy() ** -0.5) @ self.ca.svd_.V.T
        np.testing.assert_allclose(np.abs(F), np.abs(P))

    def test_total_inertia(self):
        F = robjects.r("sum(ca$eig[,1])")[0]
        P = self.ca.total_inertia_
        assert math.isclose(F, P)

    def test_eigenvalues(self):
        F = load_df_from_R("ca$eig")[: self.ca.n_components]
        P = self.ca._eigenvalues_summary
        np.testing.assert_allclose(F["eigenvalue"], P["eigenvalue"])
        np.testing.assert_allclose(F["percentage of variance"], P["% of variance"])
        np.testing.assert_allclose(
            F["cumulative percentage of variance"], P["% of variance (cumulative)"]
        )

    def test_row_coords(self, method_name="row_coordinates"):
        F = load_df_from_R(f"ca${self._row_name}$coord")
        if self.sup_rows:
            F = pd.concat((F, load_df_from_R(f"ca${self._row_name}.sup$coord")))

        method = getattr(self.ca, method_name)
        P = method(self.dataset)

        np.testing.assert_allclose(F.abs(), P.abs())

    def test_row_contrib(self):
        F = load_df_from_R(f"ca${self._row_name}$contrib")
        P = self.ca.row_contributions_
        np.testing.assert_allclose(F, P * 100)

    def test_row_cosine_similarities(self):
        F = load_df_from_R(f"ca${self._row_name}$cos2")
        if self.sup_rows:
            F = pd.concat((F, load_df_from_R(f"ca${self._row_name}.sup$cos2")))
        P = self.ca.row_cosine_similarities(self.dataset)
        np.testing.assert_allclose(F, P)

    def test_col_coords(self):
        F = load_df_from_R(f"ca${self._col_name}$coord")
        if self.sup_cols:
            F = pd.concat((F, load_df_from_R(f"ca${self._col_name}.sup$coord")))
        P = self.ca.column_coordinates(self.dataset)
        np.testing.assert_allclose(F.abs(), P.abs())

    def test_col_contrib(self):
        F = load_df_from_R(f"ca${self._col_name}$contrib")
        P = self.ca.column_contributions_
        np.testing.assert_allclose(F, P * 100)

    def test_col_cos2(self):
        F = load_df_from_R(f"ca${self._col_name}$cos2")
        if self.sup_cols:
            F = pd.concat((F, load_df_from_R(f"ca${self._col_name}.sup$cos2")))
        P = self.ca.column_cosine_similarities(self.dataset)
        np.testing.assert_allclose(F, P)



from __future__ import annotations

from . import datasets
from .ca import CA
from .famd import FAMD
from .gpa import GPA
from .mca import MCA
from .mfa import MFA
from .pca import PCA

__all__ = ["CA", "FAMD", "MCA", "MFA", "PCA", "GPA", "datasets"]



from __future__ import annotations

import numpy as np
import pytest
import rpy2.robjects as robjects
from rpy2.robjects import numpy2ri

from prince import svd
from tests import load_df_from_R


@pytest.mark.parametrize(
    "n_components, are_rows_weighted, are_columns_weighted",
    [
        pytest.param(
            n_components,
            are_rows_weighted,
            are_columns_weighted,
            id=f"{n_components=}:{are_rows_weighted=}:{are_columns_weighted=}",
        )
        for n_components in [1, 3, 10]
        for are_rows_weighted in [False, True]
        for are_columns_weighted in [False, True]
    ],
)
class TestSVD:
    @pytest.fixture(autouse=True)
    def _prepare(self, n_components, are_rows_weighted, are_columns_weighted):
        self.n_components = n_components
        self.are_rows_weighted = are_rows_weighted
        self.are_columns_weighted = are_columns_weighted

        self.dataset = np.random.rand(100, 10)
        self.row_weights = np.random.rand(100)
        self.row_weights /= self.row_weights.sum()
        self.column_weights = np.random.rand(10)

        # Fit Prince
        self.svd = svd.compute_svd(
            X=self.dataset,
            row_weights=self.row_weights if are_rows_weighted else None,
            column_weights=self.column_weights if are_columns_weighted else None,
            n_components=n_components,
            n_iter=3,
            random_state=42,
            engine="scipy",
        )

        # Fit FactoMineR
        robjects.r("library('FactoMineR')")
        robjects.r.assign("X", numpy2ri.py2rpy(self.dataset))
        robjects.r.assign("row.w", numpy2ri.py2rpy(self.row_weights))
        robjects.r.assign("col.w", numpy2ri.py2rpy(self.column_weights))
        robjects.r("row.w <- as.vector(row.w)")
        robjects.r("col.w <- as.vector(col.w)")
        args = f"X, ncp={n_components}"
        if are_rows_weighted:
            args += ", row.w=row.w"
        if are_columns_weighted:
            args += ", col.w=col.w"
        robjects.r(f"svd = svd.triplet({args})")

    def test_U(self):
        assert self.svd.U.shape == (100, self.n_components)
        if self.are_rows_weighted:
            P = self.svd.U
            F = load_df_from_R("svd$U")
            np.testing.assert_allclose(np.abs(F), np.abs(P))

    def test_s(self):
        assert self.svd.s.shape == (self.n_components,)
        if self.are_rows_weighted:
            P = self.svd.s
            F = robjects.r("svd$vs")[: self.n_components]
            np.testing.assert_allclose(np.abs(F), np.abs(P))

    def test_V(self):
        assert self.svd.V.shape == (self.n_components, 10)
        P = self.svd.V
        F = load_df_from_R("svd$V").T
        np.testing.assert_allclose(np.abs(F), np.abs(P))



"""Multiple Factor Analysis (MFA)"""

from __future__ import annotations

import collections

import altair as alt
import numpy as np
import pandas as pd

from prince import pca, utils


class MFA(pca.PCA, collections.UserDict):
    def __init__(
        self,
        n_components=2,
        n_iter=3,
        copy=True,
        check_input=True,
        random_state=None,
        engine="sklearn",
    ):
        super().__init__(
            rescale_with_mean=True,
            rescale_with_std=True,
            n_components=n_components,
            n_iter=n_iter,
            copy=copy,
            check_input=check_input,
            random_state=random_state,
            engine=engine,
        )
        collections.UserDict.__init__(self)

    @utils.check_is_dataframe_input
    def fit(self, X, y=None, groups=None, supplementary_groups=None):
        # Checks groups are provided
        self.groups_ = self._determine_groups(X, groups)
        if supplementary_groups is not None:
            for group in supplementary_groups:
                if group not in self.groups_:
                    raise ValueError(f"Supplementary group '{group}' is not in the groups")
            self.supplementary_groups_ = supplementary_groups

        # Check group types are consistent
        self.all_nums_ = {}
        for group, cols in sorted(self.groups_.items()):
            all_num = all(pd.api.types.is_numeric_dtype(X[c]) for c in cols)
            all_cat = all(pd.api.types.is_string_dtype(X[c]) for c in cols)
            if not (all_num or all_cat):
                raise ValueError(f'Not all columns in "{group}" group are of the same type')
            self.all_nums_[group] = all_num

        # Run a factor analysis in each group
        for group, cols in sorted(self.groups_.items()):
            if self.all_nums_[group]:
                fa = pca.PCA(
                    rescale_with_mean=True,
                    rescale_with_std=True,
                    n_components=self.n_components,
                    n_iter=self.n_iter,
                    copy=True,
                    random_state=self.random_state,
                    engine=self.engine,
                )
            else:
                raise NotImplementedError("Groups of non-numerical variables are not supported yet")
            self[group] = fa.fit(X.loc[:, cols])

        # Fit the global PCA
        Z = self._build_Z(X)
        column_weights = np.array(
            [
                1 / self[group].eigenvalues_[0]
                for group, cols in self.groups_.items()
                for _ in cols
                if group not in getattr(self, "supplementary_groups_", [])
            ]
        )
        super().fit(
            Z,
            column_weight=column_weights,
            supplementary_columns=[
                column
                for group in getattr(self, "supplementary_groups_", [])
                for column in self.groups_[group]
            ],
        )

        return self

    def _determine_groups(self, X: pd.DataFrame, groups: dict | list | None) -> dict:
        if groups is None:
            if isinstance(X.columns, pd.MultiIndex):
                groups = X.columns.get_level_values(0).unique().tolist()
            else:
                raise ValueError("Groups have to be specified")

        if isinstance(groups, list):
            if not isinstance(X.columns, pd.MultiIndex):
                raise ValueError(
                    "X has to have MultiIndex columns if groups are provided as a list"
                )
            groups = {
                group: [
                    (group, column)
                    for column in X.columns.get_level_values(1)[
                        X.columns.get_level_values(0) == group
                    ]
                ]
                for group in groups
            }
        return groups

    def _build_Z(self, X):
        return pd.concat(
            (X[cols] for _, cols in self.groups_.items()),
            axis="columns",
        )

    @utils.check_is_dataframe_input
    @utils.check_is_fitted
    def row_coordinates(self, X):
        """Returns the row principal coordinates."""
        Z = self._build_Z(X)
        return super().row_coordinates(Z)

    @utils.check_is_dataframe_input
    @utils.check_is_fitted
    def partial_row_coordinates(self, X):
        """Returns the partial row principal coordinates."""
        Z = self._build_Z(X)
        coords = []
        for _, names in self.groups_.items():
            partial_coords = pd.DataFrame(0.0, index=Z.index, columns=Z.columns)
            partial_coords.loc[:, names] = (Z[names] - Z[names].mean()) / Z[names].std(ddof=0)
            partial_coords = partial_coords * self.column_weight_
            partial_coords = (len(self.groups_) * partial_coords).dot(self.svd_.V.T)
            coords.append(partial_coords)
        coords = pd.concat(coords, axis=1, keys=self.groups_.keys())
        coords.columns.name = "component"
        return coords

    @utils.check_is_dataframe_input
    @utils.check_is_fitted
    def column_coordinates(self, X):
        Z = self._build_Z(X)
        return super().column_coordinates(Z)

    @utils.check_is_dataframe_input
    @utils.check_is_fitted
    def inverse_transform(self, X):
        raise NotImplementedError("MFA inherits from PCA, but this method is not implemented yet")

    @utils.check_is_dataframe_input
    @utils.check_is_fitted
    def row_standard_coordinates(self, X):
        Z = self._build_Z(X)
        return super().row_standard_coordinates(Z)

    @utils.check_is_dataframe_input
    @utils.check_is_fitted
    def row_cosine_similarities(self, X):
        Z = self._build_Z(X)
        return super().row_cosine_similarities(Z)

    @utils.check_is_dataframe_input
    @utils.check_is_fitted
    def column_cosine_similarities_(self, X):
        Z = self._build_Z(X)
        return super().column_cosine_similarities_(Z)

    @utils.check_is_dataframe_input
    @utils.check_is_fitted
    def plot(self, X, x_component=0, y_component=1, show_partial_rows=False, **params):
        index_name = X.index.name or "index"

        params["tooltip"] = (
            X.index.names if isinstance(X.index, pd.MultiIndex) else [index_name]
        ) + [
            "group",
            f"component {x_component}",
            f"component {y_component}",
        ]

        eig = self._eigenvalues_summary.to_dict(orient="index")

        row_plot = None
        partial_row_plot = None
        edges_plot = None

        # Barycenters
        row_coords = self.row_coordinates(X)
        row_coords.columns = [f"component {i}" for i in row_coords.columns]
        row_coords = row_coords.reset_index()
        row_coords["group"] = "Global"
        if show_partial_rows:
            params["color"] = "group:N"
        row_plot = (
            alt.Chart(row_coords)
            .mark_point(filled=True, size=50)
            .encode(
                alt.X(
                    f"component {x_component}",
                    scale=alt.Scale(zero=False),
                    axis=alt.Axis(
                        title=f"component {x_component} — {eig[x_component]['% of variance'] / 100:.2%}"
                    ),
                ),
                alt.Y(
                    f"component {y_component}",
                    scale=alt.Scale(zero=False),
                    axis=alt.Axis(
                        title=f"component {y_component} — {eig[y_component]['% of variance'] / 100:.2%}"
                    ),
                ),
                **params,
            )
        )

        # Partial row coordinates
        if show_partial_rows:
            partial_row_coords = self.partial_row_coordinates(X).stack(level=0, future_stack=True)
            partial_row_coords.columns = [f"component {i}" for i in partial_row_coords.columns]
            partial_row_coords = partial_row_coords.reset_index(names=[index_name, "group"])

            partial_row_plot = (
                alt.Chart(partial_row_coords)
                .mark_point(shape="circle")
                .encode(
                    alt.X(f"component {x_component}", scale=alt.Scale(zero=False)),
                    alt.Y(f"component {y_component}", scale=alt.Scale(zero=False)),
                    **params,
                )
            )

        # Edges to connect the main markers to the partial markers
        if show_partial_rows:
            edges = pd.merge(
                left=row_coords[
                    [index_name, f"component {x_component}", f"component {y_component}"]
                ],
                right=partial_row_coords[
                    [index_name, f"component {x_component}", f"component {y_component}", "group"]
                ],
                on=index_name,
                suffixes=("_global", "_partial"),
            )
            edges_plot = (
                alt.Chart(edges)
                .mark_line(opacity=0.7)
                .encode(
                    x=f"component {x_component}_global:Q",
                    y=f"component {y_component}_global:Q",
                    x2=f"component {x_component}_partial:Q",
                    y2=f"component {y_component}_partial:Q",
                    color="group:N",
                    strokeDash=alt.value([2, 2]),
                )
            )

        charts = filter(
            None,
            (row_plot, partial_row_plot, edges_plot),
        )

        return alt.layer(*charts).interactive()



"""Generalized Procrustes Analysis (GPA)"""

from __future__ import annotations

import numpy as np
from scipy.linalg import orthogonal_procrustes
from scipy.spatial import procrustes
from sklearn import base
from sklearn import utils as sk_utils

from prince import utils


class GPA(base.BaseEstimator, base.TransformerMixin):
    """Generalized Procrustes Analysis (GPA).

    Algorithm outline:

    1. Choose a reference shape.
    2. Apply Procrustes Analysis to superimpose all shapes to the reference shape.
    3. Compute the mean shape of the superimposed shapes.
    4. Repeat steps 2 and 3 until convergence.

    Parameters
    ----------
    max_iter
        The maximum number of Procrustes analysis iterations.
    tol
        The tolerance for the optimization; stops if the Procrustes distance decreases by less or
        equal to `tol` between iterations.
    init
        Method for initializing reference shape.
        - 'random' : choose reference shape from shape list
        - 'mean' : initialize reference shape as mean of shape list
    scale
        Whether to compute transformations with a scale component.
    copy
        Whether to copy data or perform the computations inplace. If False, data passed to fit are
        overwritten and running fit(X).transform(X) will not yield the expected results,
        use fit_transform(X) instead.
    check_input
        Whether to check the consistency of the inputs.
    random_state
        Determines random number generation for initialization when `init=='random'`.

    References
    ----------
    https://wikipedia.org/wiki/Generalized_Procrustes_analysis
    https://medium.com/@olga_kravchenko/generalized-procrustes-analysis-with-python-numpy-c571e8e8a421

    """

    def __init__(
        self,
        max_iter=10,
        tol=1e-4,
        init="random",
        scale=True,
        copy=True,
        check_input=True,
        random_state=None,
    ):
        self.max_iter = max_iter
        self.tol = tol
        self.init = init
        self.scale = scale
        self.copy = copy
        self.check_input = check_input
        self.random_state = random_state

    def fit(self, X, y=None):
        """Fit the model with X.

        The algorithm naturally fits and transforms at the same time, so this
        simply calls ``.fit_transform``

        Parameters:
            X (array-like of shape (n_shapes, n_points, n_dim)): Matrix of
                shapes to match to each other.
            y: Ignored

        Returns:
            self (object): The instance itself
        """
        self.fit_transform(X)

        return self

    @utils.check_is_fitted
    def transform(self, X):
        """Align X to the reference shape.

        Parameters:
            X (array-like of shape (n_shapes, n_points, n_dim)): Matrix of
                shapes to align to the refernce shape.

        Returns:
            X_new (array-like of shape (n_shapes, n_points, n_dim)): Matrix of
                aligned shapes
        """
        self._check_is_fitted()
        if self.check_input:
            self._check_input(X)

        X_new = np.empty(X.shape)
        for shape_idx in range(X.shape[0]):
            _, X_new[shape_idx], _ = procrustes(self.reference_shape, X[shape_idx])

        return X_new

    def fit_transform(self, X, y=None):
        """Fit the model with X and return the aligned shapes.

        Parameters:
            X (array-like of shape (n_shapes, n_points, n_dim)): Matrix of
                shapes to match to each other.
            y: Ignored

        Returns:
            X_new (array-like of shape (n_shapes, n_points, n_dim)): Matrix X
                of aligned shapes
        """

        # Check input
        if self.check_input:
            self._check_input(X)

        # Copy data
        if self.copy:
            X = np.array(X, copy=True)

        # scikit-learn SLEP010
        n_shapes, n_points, n_dim = X.shape
        self.n_features_in_ = n_dim

        # Pick reference shape
        if self.init == "random":
            random_state = sk_utils.check_random_state(self.random_state)
            ref_shape_idx = random_state.randint(X.shape[0])
            reference_shape = X[ref_shape_idx].copy()
        elif self.init == "mean":
            reference_shape = X.mean(axis=0)
        else:
            raise ValueError("init method must be one of ('random', 'mean')")

        for iter_idx in range(self.max_iter):
            # Align each shape to reference shape
            for shape_idx in range(X.shape[0]):
                if self.scale:
                    _, X[shape_idx], _ = procrustes(reference_shape, X[shape_idx])
                else:
                    _, X[shape_idx] = unscaled_procrustes(reference_shape, X[shape_idx])

            # Compute diagnostics
            mean_shape = X.mean(axis=0)
            procrustes_distance = np.linalg.norm(reference_shape - mean_shape)

            # Update reference shape
            reference_shape = mean_shape

            # Check for convergence
            if procrustes_distance <= self.tol:
                break

        # Store properties
        self._reference_shape = reference_shape

        # Return the aligned shapes
        return X

    def _check_input(self, X):
        sk_utils.check_array(X, allow_nd=True)
        if X.ndim != 3:
            raise ValueError("Expected 3-dimensional input of (n_shapes, n_points, n_dim)")

    def _check_is_fitted(self):
        sk_utils.validation.check_is_fitted(self, "_reference_shape")

    @property
    def reference_shape(self):
        """Returns the final reference shape."""
        self._check_is_fitted()
        return self._reference_shape


def unscaled_procrustes(reference, data):
    """Fit `data` to `reference` using procrustes analysis without scaling.
    Uses translation (mean-centering), reflection, and orthogonal rotation.

    Parameters:
        reference (array-like of shape (n_points, n_dim)): reference shape to
            fit `data` to
        data (array-like of shape (n_points, n_dim)): shape to align to
            `reference`

    Returns:
        reference_centered (np.ndarray of shape (n_points, n_dim)): 0-centered
            `reference` shape
        data_aligned (np.ndarray of shape (n_points, n_dim)): `data` aligned to
            the reference shape
    """
    # Convert inputs to np.ndarray types
    reference = np.array(reference, dtype=np.double)
    data = np.array(data, dtype=np.double)

    # Translate data to the origin
    reference_centered = reference - reference.mean(axis=0)
    data_centered = data - data.mean(axis=0)

    # Rotate / reflect data to match reference
    # transform mtx2 to minimize disparity
    R, _ = orthogonal_procrustes(data_centered, reference_centered)
    data_aligned = data_centered @ R

    return reference_centered, data_aligned
